@@ -1,7 +1,8 @@
-use crate::Pid;
+use crate::{Error, Pid};
 use ferrisetw::native::etw_types::EventRecord;
 use ferrisetw::provider::Provider;
 use ferrisetw::trace::{TraceBaseTrait, TraceTrait, UserTrace};
+use ntapi::ntpsapi::{NtQueryInformationProcess, ProcessVmCounters, VM_COUNTERS_EX2};
 use serde::Deserialize;
 use std::collections::HashMap;
 use std::io::BufReader;
@@ -13,9 +14,13 @@ use std::sync::{Arc, RwLock};
 use std::thread;
 use std::time::{Duration, Instant};
 use std::{mem, process};
-use winapi::shared::minwindef::DWORD;
+use winapi::shared::minwindef::{DWORD, FALSE};
+use winapi::shared::ntdef::NT_SUCCESS;
 use winapi::shared::winerror::ERROR_SUCCESS;
+use winapi::um::handleapi::CloseHandle;
 use winapi::um::pdh::*;
+use winapi::um::processthreadsapi::OpenProcess;
+use winapi::um::winnt::{HANDLE, PROCESS_QUERY_INFORMATION, PROCESS_VM_READ};
 
 // https://docs.microsoft.com/en-us/windows/win32/perfctrs/pdh-error-codes
 // 0x800007D2 (PDH_MORE_DATA)
@@ -129,55 +134,61 @@ pub struct Pdh {
 }
 
 impl Pdh {
-    pub fn new<T: IntoIterator<Item = Pid>>(pids: T) -> Self {
+    pub fn new<T: IntoIterator<Item = Pid>>(pids: T) -> Result<Self, Error> {
         unsafe {
             let mut query = MaybeUninit::uninit().assume_init();
             let mut r = PdhOpenQueryW(ptr::null(), 0, &mut query);
-            assert_eq!(r, ERROR_SUCCESS as _);
+            if r != ERROR_SUCCESS as PDH_STATUS {
+                return Err(Error::Pdh(r));
+            }
 
-            let process_gpu_counters = pids
-                .into_iter()
-                .map(|pid| {
-                    let mut process_gpu_counter: PDH_HCOUNTER = MaybeUninit::uninit().assume_init();
-                    r = PdhAddCounterW(
-                        query,
-                        widestring::U16CString::from_str(format!(
-                            "\\GPU Engine(pid_{}*)\\Utilization Percentage",
-                            pid
-                        ))
-                        .unwrap()
-                        .as_ptr(),
-                        0,
-                        &mut process_gpu_counter,
-                    );
-                    assert_eq!(r, ERROR_SUCCESS as _);
+            let mut pdh = Self {
+                query,
+                process_gpu_counters: vec![],
+                total_gpu_counter: MaybeUninit::uninit().assume_init(),
+            };
 
-                    ProcessCounter {
-                        pid,
-                        counter: process_gpu_counter,
-                    }
-                })
-                .collect();
+            for pid in pids {
+                let mut process_gpu_counter: PDH_HCOUNTER = MaybeUninit::uninit().assume_init();
+                r = PdhAddCounterW(
+                    query,
+                    widestring::U16CString::from_str(format!(
+                        "\\GPU Engine(pid_{}*)\\Utilization Percentage",
+                        pid
+                    ))
+                    .unwrap()
+                    .as_ptr(),
+                    0,
+                    &mut process_gpu_counter,
+                );
+                if r != ERROR_SUCCESS as PDH_STATUS {
+                    return Err(Error::Pdh(r));
+                }
+
+                pdh.process_gpu_counters.push(ProcessCounter {
+                    pid,
+                    counter: process_gpu_counter,
+                });
+            }
 
-            let mut total_gpu_counter: PDH_HCOUNTER = MaybeUninit::uninit().assume_init();
             r = PdhAddCounterW(
                 query,
                 widestring::U16CString::from_str("\\GPU Engine(*)\\Utilization Percentage")
                     .unwrap()
                     .as_ptr(),
                 0,
-                &mut total_gpu_counter,
+                &mut pdh.total_gpu_counter,
             );
-            assert_eq!(r, ERROR_SUCCESS as _);
+            if r != ERROR_SUCCESS as PDH_STATUS {
+                return Err(Error::Pdh(r));
+            }
 
             r = PdhCollectQueryData(query);
-            assert_eq!(r, ERROR_SUCCESS as _);
-
-            Self {
-                query,
-                process_gpu_counters,
-                total_gpu_counter,
+            if r != ERROR_SUCCESS as PDH_STATUS {
+                return Err(Error::Pdh(r));
             }
+
+            Ok(pdh)
         }
     }
 
@@ -391,3 +402,67 @@ struct PresentInfo {
     last_time: Instant,
     count: u32,
 }
+
+pub struct VmCounter {
+    process_counters: Vec<ProcessVmCounter>,
+}
+
+impl VmCounter {
+    pub fn new<T: IntoIterator<Item = Pid>>(pids: T) -> Result<Self, Error> {
+        let mut vm_counter = Self {
+            process_counters: vec![],
+        };
+
+        for pid in pids {
+            let options = PROCESS_QUERY_INFORMATION | PROCESS_VM_READ;
+            let handle = unsafe { OpenProcess(options, FALSE, pid as DWORD) };
+            if handle.is_null() {
+                return Err(Error::ProcessHandle);
+            } else {
+                vm_counter.process_counters.push(ProcessVmCounter {
+                    pid,
+                    handle,
+                    mem: 0.0,
+                });
+            }
+        }
+
+        Ok(vm_counter)
+    }
+
+    pub fn process_mem(&self, pid: Pid) -> Option<f32> {
+        self.process_counters
+            .iter()
+            .find(|p| p.pid == pid)
+            .map(|p| unsafe {
+                let mut info: VM_COUNTERS_EX2 = MaybeUninit::uninit().assume_init();
+                let r = NtQueryInformationProcess(
+                    p.handle,
+                    ProcessVmCounters,
+                    std::mem::transmute(&mut info),
+                    std::mem::size_of::<VM_COUNTERS_EX2>() as _,
+                    std::ptr::null_mut(),
+                );
+                if NT_SUCCESS(r) {
+                    Some((info.PrivateWorkingSetSize >> 10) as f32)
+                } else {
+                    None
+                }
+            })
+            .flatten()
+    }
+}
+
+struct ProcessVmCounter {
+    pid: Pid,
+    handle: HANDLE,
+    mem: f32,
+}
+
+impl Drop for ProcessVmCounter {
+    fn drop(&mut self) {
+        unsafe {
+            let _ = CloseHandle(self.handle);
+        }
+    }
+}
